#!/usr/bin/perl -w
#################################################################################
# Prepare data about the given component to be inserted into the database.
#
# This script uses the output of the system package manager (rpm, dpkg or
# emerge), analyzing the files that are provided by the component.
#
# The script creates a text containing list of components, libraries and
# interfaces (actually for each shared object and executable the output of
# 'readelf' is placed into the output file; see README_collect for details).
# If the script is called for a standalone component then the output file should
# be processed by the 'upload_distr_data.pl' script on the server side for actual
# data insertion. In all other cases files produced by componenttodb.pl will
# be joined by apptodb.pl or distrtodb.pl and resultin files will be processed
# by appropriate upload* scripts.
#
# Classes will be detected while parsing interface names in the 'upload_data.pl'
#
# Symlinks and files that are not executable or shared objects are ignored.
#
# For shared objects, the script tries to obtain ABItag using either ldconfig
# cache or objdump
#
# Author: Denis Silakov <silakov@ispras.ru>
#################################################################################

use strict;
use Getopt::Long;

use Text::Wrap;
Getopt::Long::Configure ("posix_default", "no_ignore_case");

use Distros;

# Component information
my $ComponentName;
my $ComponentVersion="";
my $ComponentId=0;
my $ComponentArch;
my $ComponentDbName;

# Package manager used in the distribution
my $PackageManager="";

# All about rpms
my @rpms;     # rpms that provide the component
my @sofiles;  # list of files from rpms that provide the component

# All about library under process
my $LibName;
my $LibRunName = "";
my $LibSoName = "";
my $FileName = "";
my $LibPath = "";

# Options
my $quiet = 0;
my $opt_help;
my $OutputFile;
my $ComponentFilesList;

GetOptions("help|h!" => \$opt_help,
            "quiet|q!" => \$quiet,
            "arch|a=s" => \$ComponentArch,
            "packagemanager|m=s" => \$PackageManager,
            "comp-dbname|c=s" => \$ComponentDbName,
            "version|v=s" => \$ComponentVersion,
            "filelist|l=s" => \$ComponentFilesList,
            "file|f=s" => \$OutputFile);

sub HELP_MESSAGE();
sub add_component();
sub add_library();

# Functions for obtaining information using different package managers
sub get_info_rpm();
sub get_info_dpkg();
sub get_info_emerge();
sub get_info_by_filelist();

HELP_MESSAGE() if (defined $opt_help);

if(@ARGV != 1) {HELP_MESSAGE(); exit(1);}

# Guess package manager, if it was not specified
if( (not $PackageManager) and (not $ComponentFilesList) ) {
    $PackageManager = guess_package_manager();
    if( $PackageManager eq "unknown" ) {
        print STDERR "Cannot guess package manager! Please specify it directly using '-p' option.";
        exit(1);
    }
    else {
        print STDERR "No package manager was specified, guessing it is $PackageManager\n";
    }
}

$ComponentName = $ARGV[0];

if( not $ComponentDbName ) {
    $ComponentDbName = $ComponentName;
}

if( defined $OutputFile ) {
    open( COMPONENT_DATA, "> \"$OutputFile\"" );
}
else {
    open( COMPONENT_DATA, "> \"".$ComponentName."_upload_data\"" );
	$OutputFile = $ComponentName."_upload_data";
}

if( not $ComponentArch ) {
    $ComponentArch = guess_platform(`uname -m`);
}

##############################################################################
# Main section
##############################################################################

if( $ComponentFilesList ) {
	get_info_by_filelist();
}
elsif( $PackageManager eq "rpm" ) {
    get_info_rpm();
}
elsif( $PackageManager eq "dpkg" ) {
    get_info_dpkg();
}
elsif( $PackageManager eq "emerge" ) {
    if( does_shell_know("equery") ) {
        get_info_emerge();
    }
    else {
        print STDERR "This script requires 'equery' tool from 'gentoolkit' package.\n";
        exit(1);
    }
}
else {
    print STDERR "Package manager '$PackageManager' is not supproted\n";
    exit(1);
}

# Check the types of every file and call appropriate function
for( 0..$#sofiles )
{
    my $soname = $sofiles[$_];

    print "Processing file: '$soname'\n" if(not $quiet);

    # Skip symlinks
    if( -l($soname) ) {
        print "    Symlink, skipping...\n" if(not $quiet);
        next;
    }

    my $basename = `basename '$soname'`;
    chomp($basename);

    # What does 'file' think about our file -
    #  is it an executable or a shared object?
    my $type = `file '$soname'`;

	if( $type =~ m/\:\s*(\w(\s|\:|\-|\/|\w|\+|\.)*)\s+script/ ) {
        print "Found a script: $soname\n" if(not $quiet);
        print COMPONENT_DATA "!Script '$basename' '$soname' '$1'\n";
		next;
	}
    elsif( $type =~ m/MS-DOS executable/ ) {
        print "Found MS-DOS executable: $soname\n" if(not $quiet);
        print COMPONENT_DATA "!Command '$basename' '$soname'\n";
		next;
    }
    elsif( $type =~ m/executable/ ) {
        print "Found a command: $soname\n" if(not $quiet);
        print COMPONENT_DATA "!Command '$basename' '$soname'\n";

		system "readelf -Wa \"$soname\" | LC_ALL=C grep -v R_X86_64_GLOB_DAT | LC_ALL=C grep -v R_X86_64_COPY | LC_ALL=C grep -v R_X86_64_JUMP_SLOT | LC_ALL=C grep -v R_386_32 | LC_ALL=C grep -v R_386_COPY | LC_ALL=C grep -v R_386_GLOB_DAT | LC_ALL=C grep -v R_386_JUMP_SLOT | LC_ALL=C grep -v R_386_RELATIVE | LC_ALL=C grep -v R_IA64_REL64LSB | LC_ALL=C grep -v R_IA64_DIR64LSB | LC_ALL=C grep -v R_IA64_FPTR64LSB | LC_ALL=C grep -v R_X86_64_RELATIVE > /var/tmp/componenttodb_lib";
		close( COMPONENT_DATA );
		system "cat /var/tmp/componenttodb_lib >> \"$OutputFile\"";
		open( COMPONENT_DATA, ">> $OutputFile" );
		system "rm -f /var/tmp/componenttodb_lib";
# 		my $elfdata = `readelf -Wa \"$soname\"`;
# 		print COMPONENT_DATA "$elfdata\n";
		next;
    }
    elsif( $type =~ m/shared object/ ) {
    	# Sometimes commands are also reported to be 'shared objects' by file...
    	# Try to check with 'extract', if it is present
#     	if( does_shell_know("extract") ) {
#     		my $extract_result = `extract $soname | grep -i shared\\ object`;
#     		if( !$extract_result ) {
#     			# Seems to be a command...
# 				print "Found a command: $soname\n" if(not $quiet);
# 				print "Found a command: $soname ($extract_result)\n";
# 				print COMPONENT_DATA "!Command '$basename' '$soname'\n";
#
# 				my $elfdata = `readelf -Wa \"$soname\"`;
# 				print COMPONENT_DATA "$elfdata\n";
# 				next;
#     		}
#     	}
        print "Found a shared object: $soname\n" if(not $quiet);

        $LibRunName = $basename;
        $LibPath = $soname;

        $LibPath =~ s/\/\Q$basename\E//;

        $FileName = $soname;
        ($LibName, my @temp) = split /\.so/, $LibRunName;

        # Add library data
        add_library();

		system "readelf -Wa \"$soname\" | LC_ALL=C grep -v R_X86_64_GLOB_DAT | LC_ALL=C grep -v R_X86_64_COPY | LC_ALL=C grep -v R_X86_64_JUMP_SLOT | LC_ALL=C grep -v R_386_32 | LC_ALL=C grep -v R_386_COPY | LC_ALL=C grep -v R_386_GLOB_DAT | LC_ALL=C grep -v R_386_JUMP_SLOT | LC_ALL=C grep -v R_386_RELATIVE | LC_ALL=C grep -v R_IA64_REL64LSB | LC_ALL=C grep -v R_IA64_DIR64LSB | LC_ALL=C grep -v R_IA64_FPTR64LSB | LC_ALL=C grep -v R_X86_64_RELATIVE > /var/tmp/componenttodb_lib";
		close( COMPONENT_DATA );
		system "cat /var/tmp/componenttodb_lib >> \"$OutputFile\"";
		open( COMPONENT_DATA, ">> $OutputFile" );
		system "rm -f /var/tmp/componenttodb_lib";
		#my $elfdata = `readelf -Wa \"$soname\"`;
		#print COMPONENT_DATA "$elfdata\n";
    }
    else {
        print "    Not executable or shared object, skipping...\n" if( not $quiet );
    }
}

################################################################################################
# Collect information about components directly by file list or using different package managers
################################################################################################

sub get_info_by_filelist()
{
	open( FILE_LIST, "< $ComponentFilesList" ) || die "Cannot open file with filelist";

	my $tmpName = $ComponentName;
	$ComponentName = $ComponentDbName;
	add_component();
	$ComponentName = $tmpName;

	my $type;
    my $j=0;

	while( <FILE_LIST> ) {
		chomp();
		$type = `file '$_'`;

		if( $type =~ /directory/ ) {
			open( FILES, "find '$_' -type f -print |" ) || die "Cannot process files from directory $_";

			while( <FILES> ) {
				chomp;
				$sofiles[$j] = $_;
				$j++;
			}

			close FILES;
		}
		else {
			$sofiles[$j] = $_;
			$j++;
		}
	}

	close FILE_LIST;
}

sub get_info_rpm()
{
    # Get a list of rpm that provide the given component
    open(RPM_PROV,"rpm -q --whatprovides $ComponentName|") || die "Cannot get information from rpm.";

    @rpms = <RPM_PROV>;

    chomp($rpms[0]);
    if( $rpms[0] eq "no package provides $ComponentName" ) {
		# Maybe we are processing an application and 'whatprovides' doesn't
		# know about it?
		# (If we are wrong here and $ComponentName is not a valid rpm name,
		#  we'll not obtain the file list - nothing wrong, but user should be informed)
		$rpms[0] = $ComponentName;
    }

    close RPM_PROV;

    # If component version is not set manually, then take it from the first rpm
	if( !$ComponentVersion ) {
		if ($rpms[0] =~ /((\w|\d|\-\w)*)\-((\w|\d|\.|\-)*)/ ) {
			$ComponentVersion = $3;
		}
		else {
			print STDERR "$ComponentName: Cannot get component version!\n";
			$ComponentVersion = "";
		}
	}

    add_component();

    my $j=0;

    # Explore files provided by rmps
    for( 0..$#rpms ) {
        chomp($rpms[$_]);
        open(FILE_PROV,"rpm -q --fileprovide ".$rpms[$_]." | ") || die "Cannot get information from rpm.";

        while(<FILE_PROV>) {
            chomp;

			if( /not installed / ) {
				print STDERR "Missing package!\n";
				next;
			}

            if( /((\/(\w|\.|\-|\+|\[)+)*)\s*((\w+\/\s*)+)*/ ) {
                print "file found: '".$1."'\n" if(not $quiet);
                $sofiles[$j] = $1;
                $j++;
            }

        }

        close FILE_PROV;
    }
}

sub get_info_dpkg()
{
    # Unfortunately, we cannot use something like 'rpm -q --whatprovides' here,
    # cause for many base components (like Qt, Gtk etc.) there are no
    # packages that 'Provide' them. So, user should provide us with correct
    # package name

    # Check if the component is available
    my $notAvailable = `dpkg -s "$ComponentName" | grep "is not installed"`;
    if( $notAvailable ) {
        print STDERR "Package $ComponentName is not installed and cannot be processed.\n";
        return;
    }

    # If component version is not set manually, then take it from the first package
	if( !$ComponentVersion ) {
		my $ver = `dpkg -s "$ComponentName" | LC_ALL=C grep -i version`;
		if( $ver =~ /(V|v)ersion\s*:\s*(\S+)/ ) {
			$ComponentVersion = $2;
		}
		else {
			print STDERR "$ComponentName: Cannot get component version!\n";
			$ComponentVersion = "";
		}
	}

    add_component();

    my $j=0;

    open(FILE_PROV,"dpkg -L \"$ComponentName\" | ") || die "Cannot get information from dpkg.";
    while(<FILE_PROV>) {
        chomp;
        print "file found: '".$_."'\n" if(not $quiet);
        $sofiles[$j] = $_;
        $j++;
    }

    close FILE_PROV;
}

sub get_info_emerge()
{
    # Get component version, if it is not specified directly

    if( !$ComponentVersion ) {
        my $VerLine = `equery which "$ComponentName"`;
        if( $VerLine =~ /\Q$ComponentName\E-(.+)\.ebuild/ ) {
            $ComponentVersion=$1;
        }
        else {
            print STDERR "Cannot get component version for '$ComponentName'!\n";
        }
    }

    add_component();

    my $j=0;

    open(FILE_PROV,"equery files $ComponentName | ") || die "Cannot get information from equery.";
    while(<FILE_PROV>) {
        chomp;
        print "file found: '".$_."'\n" if(not $quiet);
        $sofiles[$j] = $_;
        $j++;
    }

    close FILE_PROV;
}

##############################################################################
# Functions for printing additional information about artifacts
##############################################################################
#
# add_library: print record with library information
#
sub add_library()
{
	# Get soname of the library
	$LibSoName='';
	my $SonameString='';
	$SonameString = `readelf -d \"$LibPath/$LibRunName\" | LC_ALL=C fgrep SONAME`;
	if( $SonameString =~ /Library soname: \[((\S)*)\]/ ) {
		$LibSoName = $1;
	}

	my $ABItag='';

	# Try to obtain ABI tag from ldconfig cache
	my $TagString = `/sbin/ldconfig -p | LC_ALL=C fgrep "$LibPath/$LibSoName"`;
	if( $TagString =~ /OS ABI\: ((\w|\d| |\.|\-|\:)*)\)/ ) {
		$ABItag = $1;
	}

	if( not $ABItag ) {
		# Try to obtain ABI tag using objdump
		open(OBJDUMP,"objdump -s -j .note.ABI-tag \"$LibPath/$LibRunName\"|") || die "Cannot get information from objdump.";

		my $found=0;

		my @dump = <OBJDUMP>;
		for( 0..$#dump )
		{
			chomp($dump[$_]);

			# We should achieve two lines of bytes; 21st, 25th and 29th bytes
			# form ABI version
			if( $dump[$_] =~ /\d\d\d\d/ ) {
				if( $found == 1 ) {
					if( $dump[$_] =~ /\d{4} (\d{8}) (\d{2})\d{6} (\d{2})\d{6} (\d{2})\d{6}/ ) {
						(my $name, my $a, my $b, my $c) = ($1,$2,$3,$4);

						if( $a =~ /^0(\d)/ ) {
							$a=$1;
						}
						if( $b =~ /^0(\d)/ ) {
							$b=$1;
						}
						if( $c =~ /^0(\d)/ ) {
							$c=$1;
						}

						if( $name ne "Linux" ) {
							if( $name == 0 ) {
								$ABItag = "Linux $a.$b.$c";
							}
							else {
								$ABItag = "$name $a.$b.$c";
							}
						}
						else {
							$ABItag = "$name $a.$b.$c";
						}
					}
					$found++;
				}
				else {
					$found++;
				}
			}
		}
	}

	# Try to detect library version (quite difficult task...)
	if( ($LibName =~ /^ISO/) or ($LibName =~ /^GOST/) or ($LibName =~ /^UTF/) or ($LibName =~ /^TIS/)
			or ($LibName =~ /^ISIRI/) ) {
    	print COMPONENT_DATA "!Library '$LibName' '' '$LibRunName' '$LibPath' '$ABItag' '$LibSoName'\n";
	}
	elsif( $LibName =~ /(lib(g|stdc)\+\+\-(\d-)?libc\d)\.(.+)/ ) {
		my $lname = $1;
		my $lver = $4;
    	print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath' '$ABItag' '$LibSoName'\n";
	}
	elsif( $LibName =~ /((\w|\d|_|\+|\-\w)+)\-((\d).+)/ ) {
		my $lname = $1;
		my $lver = $3;
    	print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath' '$ABItag' '$LibSoName'\n";
	}
	elsif( $LibRunName =~ /((\w|\d|_|\+|\.|\-\w)+)\.so\.((\d).+)/ ) {
		my $lname = $1;
		my $lver = $3;
    	print COMPONENT_DATA "!Library '$lname' '$lver' '$LibRunName' '$LibPath' '$ABItag' '$LibSoName'\n";
	}
	else {
    	print COMPONENT_DATA "!Library '$LibName' '' '$LibRunName' '$LibPath' '$ABItag' '$LibSoName'\n";
	}
}

sub add_component()
{
    # Remove the leading '=', if presents, in case of Gentoo portage
    if( $PackageManager eq "emerge" ) {
        my $PackageName = $ComponentName;
        $PackageName =~ s/^\=//;
	if( $PackageName =~ /(-$ComponentVersion)$/ ) {
            $PackageName =~ s/-$ComponentVersion//;
	}
        print COMPONENT_DATA "!Component '$PackageName' '$ComponentVersion' '$ComponentDbName' '$ComponentArch'\n";
    }
    else {
        print COMPONENT_DATA "!Component '$ComponentName' '$ComponentVersion' '$ComponentDbName' '$ComponentArch'\n";
    }
}

##############################################################################
# Miscellaneous functions
##############################################################################

sub HELP_MESSAGE()
{
print STDERR <<"EOM"

$0: Obtain information about libraries, commands and interfaces provided by the given component.

Usage: $0 [options] component_name

component_name                Name of the component to process as it should be
                              given to the package manager
-c|--comp-dbname              Name of the component that should be placed into the database
                              (by default <component_name> is used)
-h|--help                     Display this help message
-f|--file filename            Filename where the data for upload will be situated
                              By default, "<component_name>_upload_data" is used
-l|--filelist filename        File containing list of file names that are installed by the application.
                              (Directory names are also allowed; all files from given directories will be processed)
-v|--version version          Component version. By default, the script will try to obtain version using
                              package manager, but that version might be not that one you'd like to see if
                              the component is provided by several packages, since each package may have
                              its own version
-a|--arch                     Architecture name where the component is processed
                              (if it is not set, the output of 'uname' will be used)
-m|--packagemanager pkg_name  Package manager used in the distribution. Supported values
                              are 'rpm', 'dpkg' and 'emerge'. If none is specified, the script
                              will try to guess it.
-q|--quiet                    Do not print any information messages

EOM
      ; exit(1);
}
